/*
 * SYNCHRONY - Modified 2006/09/15 by Maurice Blount
 * based on the program
 *
 * synergy -- mouse and keyboard sharing utility
 * Copyright (C) 2002 Chris Schoeneman
 * 
 * This package is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * found in the file COPYING that should have accompanied this file.
 * 
 * This package is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#include "CClient.h"
#include "CServerProxy.h"
#include "CScreen.h"
#include "CPacketStreamFilter.h"
#include "CProtocolUtil.h"
#include "IPlatformScreen.h"
#include "ProtocolTypes.h"
#include "XSynergy.h"
#include "IDataSocket.h"
#include "ISocketFactory.h"
#include "IStreamFilterFactory.h"
#include "CLog.h"
#include "IEventQueue.h"
#include "TMethodEventJob.h"


//
// CClient
//

CEvent::Type			CClient::s_connectedEvent        = CEvent::kUnknown;
CEvent::Type			CClient::s_connectionFailedEvent = CEvent::kUnknown;
CEvent::Type			CClient::s_disconnectedEvent     = CEvent::kUnknown;

CClient::CClient(const CStringSYN& name, const CNetworkAddress& address,
				ISocketFactory* socketFactory,
				IStreamFilterFactory* streamFilterFactory,
				CScreen* screen) :
	m_name(name),
	m_serverAddress(address),
	m_socketFactory(socketFactory),
	m_streamFilterFactory(streamFilterFactory),
	m_screen(screen),
	m_stream(NULL),
	m_timer(NULL),
	m_server(NULL),
	m_ready(false),
	m_active(false),
	m_suspended(false),
	m_connectOnResume(false),
	m_isInJumpzone(false),
	m_isTwoTapActive(0)
{
	assert(m_socketFactory != NULL);
	assert(m_screen        != NULL);

	// register suspend/resume event handlers
	EVENTQUEUE->adoptHandler(IScreen::getSuspendEvent(),
							getEventTarget(),
							new TMethodEventJob<CClient>(this,
								&CClient::handleSuspend));
	EVENTQUEUE->adoptHandler(IScreen::getResumeEvent(),
							getEventTarget(),
							new TMethodEventJob<CClient>(this,
								&CClient::handleResume));
}

CClient::~CClient()
{
	EVENTQUEUE->removeHandler(IScreen::getSuspendEvent(),
							  getEventTarget());
	EVENTQUEUE->removeHandler(IScreen::getResumeEvent(),
							  getEventTarget());

	cleanupTimer();
	cleanupScreen();
	cleanupConnecting();
	cleanupConnection();
	delete m_socketFactory;
	delete m_streamFilterFactory;
}

void
CClient::connect()
{
	if (m_stream != NULL) {
		return;
	}
	if (m_suspended) {
		m_connectOnResume = true;
		return;
	}

	try {
		// resolve the server hostname.  do this every time we connect
		// in case we couldn't resolve the address earlier or the address
		// has changed (which can happen frequently if this is a laptop
		// being shuttled between various networks).  patch by Brent
		// Priddy.
		m_serverAddress.resolve();

		// create the socket
		IDataSocket* socket = m_socketFactory->create();

		// filter socket messages, including a packetizing filter
		m_stream = socket;
		if (m_streamFilterFactory != NULL) {
			m_stream = m_streamFilterFactory->create(m_stream, true);
		}
		m_stream = new CPacketStreamFilter(m_stream, true);

		// connect
		LOG((CLOG_DEBUG1 "connecting to server"));
		setupConnecting();
		setupTimer();
		socket->connect(m_serverAddress);
	}
	catch (XBase& e) {
		cleanupTimer();
		cleanupConnecting();
		delete m_stream;
		m_stream = NULL;
		LOG((CLOG_DEBUG1 "connection failed"));
		sendConnectionFailedEvent(e.what());
		return;
	}
}

void
CClient::disconnect(const char* msg)
{
	m_connectOnResume = false;
	cleanupTimer();
	cleanupScreen();
	cleanupConnecting();
	cleanupConnection();
	if (msg != NULL) {
		sendConnectionFailedEvent(msg);
	}
	else {
		sendEvent(getDisconnectedEvent(), NULL);
	}
}

void
CClient::handshakeComplete()
{
	m_ready = true;
	m_screen->enable();
	sendEvent(getConnectedEvent(), NULL);
}

bool
CClient::isConnected() const
{
	return (m_server != NULL);
}

bool
CClient::isConnecting() const
{
	return (m_timer != NULL);
}

CNetworkAddress
CClient::getServerAddress() const
{
	return m_serverAddress;
}

CEvent::Type
CClient::getConnectedEvent()
{
	return CEvent::registerTypeOnce(s_connectedEvent,
							"CClient::connected");
}

CEvent::Type
CClient::getConnectionFailedEvent()
{
	return CEvent::registerTypeOnce(s_connectionFailedEvent,
							"CClient::failed");
}

CEvent::Type
CClient::getDisconnectedEvent()
{
	return CEvent::registerTypeOnce(s_disconnectedEvent,
							"CClient::disconnected");
}

void*
CClient::getEventTarget() const
{
	return m_screen->getEventTarget();
}

void
CClient::getShape(SInt32& x, SInt32& y, SInt32& w, SInt32& h) const
{
	m_screen->getShape(x, y, w, h);
}

void
CClient::getCursorPos(SInt32& x, SInt32& y) const
{
	m_screen->getCursorPos(x, y);
}

void
CClient::enter(SInt32 xAbs, SInt32 yAbs)
{
	m_active = true;
	m_isInJumpzone = false;
	m_screen->warpCursor(xAbs, yAbs);
	m_screen->enter();
}

bool
CClient::leave()
{
	// prevent rogue mouse motions interfering with switch
	lockMouseTimer.reset();
	lockMouseTimer.start();

	// leave screen
	m_screen->leave();
	m_active = false;

	// tell server we have left so it will accept any input
	CProtocolUtil::writef(m_stream, kMsgCHasLeft);
	return true;
}

void
CClient::resetOptions()
{
	m_screen->resetOptions();
}

void
CClient::setOptions(const COptionsList& options)
{
	m_screen->setOptions(options);
}

CStringSYN
CClient::getName() const
{
	return m_name;
}

void
CClient::sendEvent(CEvent::Type type, void* data)
{
	EVENTQUEUE->addEvent(CEvent(type, getEventTarget(), data));
}

void
CClient::sendConnectionFailedEvent(const char* msg)
{
	CFailInfo* info = (CFailInfo*)malloc(sizeof(CFailInfo) + strlen(msg));
	info->m_retry   = true;
	strcpy(info->m_what, msg);
	sendEvent(getConnectionFailedEvent(), info);
}

void
CClient::setupConnecting()
{
	assert(m_stream != NULL);

	EVENTQUEUE->adoptHandler(IDataSocket::getConnectedEvent(),
							m_stream->getEventTarget(),
							new TMethodEventJob<CClient>(this,
								&CClient::handleConnected));
	EVENTQUEUE->adoptHandler(IDataSocket::getConnectionFailedEvent(),
							m_stream->getEventTarget(),
							new TMethodEventJob<CClient>(this,
								&CClient::handleConnectionFailed));
}

void
CClient::setupConnection()
{
	assert(m_stream != NULL);

	EVENTQUEUE->adoptHandler(ISocket::getDisconnectedEvent(),
							m_stream->getEventTarget(),
							new TMethodEventJob<CClient>(this,
								&CClient::handleDisconnected));
	EVENTQUEUE->adoptHandler(IStreamSYN::getInputReadyEvent(),
							m_stream->getEventTarget(),
							new TMethodEventJob<CClient>(this,
								&CClient::handleHello));
	EVENTQUEUE->adoptHandler(IStreamSYN::getOutputErrorEvent(),
							m_stream->getEventTarget(),
							new TMethodEventJob<CClient>(this,
								&CClient::handleOutputError));
	EVENTQUEUE->adoptHandler(IStreamSYN::getInputShutdownEvent(),
							m_stream->getEventTarget(),
							new TMethodEventJob<CClient>(this,
								&CClient::handleDisconnected));
	EVENTQUEUE->adoptHandler(IStreamSYN::getOutputShutdownEvent(),
							m_stream->getEventTarget(),
							new TMethodEventJob<CClient>(this,
								&CClient::handleDisconnected));
}

void
CClient::setupScreen()
{
	assert(m_server == NULL);

	m_ready  = false;
	m_server = new CServerProxy(this, m_stream);
	EVENTQUEUE->adoptHandler(IScreen::getShapeChangedEvent(),
							getEventTarget(),
							new TMethodEventJob<CClient>(this,
								&CClient::handleShapeChanged));
	EVENTQUEUE->adoptHandler(IPlatformScreen::getMotionEvent(),
							getEventTarget(),
							new TMethodEventJob<CClient>(this,
								&CClient::handleMotionEvent));
}

void
CClient::setupTimer()
{
	assert(m_timer == NULL);

	m_timer = EVENTQUEUE->newOneShotTimer(5.0, NULL);
	EVENTQUEUE->adoptHandler(CEvent::kTimer, m_timer,
							new TMethodEventJob<CClient>(this,
								&CClient::handleConnectTimeout));
}

void
CClient::cleanupConnecting()
{
	if (m_stream != NULL) {
		EVENTQUEUE->removeHandler(IDataSocket::getConnectedEvent(),
							m_stream->getEventTarget());
		EVENTQUEUE->removeHandler(IDataSocket::getConnectionFailedEvent(),
							m_stream->getEventTarget());
	}
}

void
CClient::cleanupConnection()
{
	if (m_stream != NULL) {
		EVENTQUEUE->removeHandler(IStreamSYN::getInputReadyEvent(),
							m_stream->getEventTarget());
		EVENTQUEUE->removeHandler(IStreamSYN::getOutputErrorEvent(),
							m_stream->getEventTarget());
		EVENTQUEUE->removeHandler(IStreamSYN::getInputShutdownEvent(),
							m_stream->getEventTarget());
		EVENTQUEUE->removeHandler(IStreamSYN::getOutputShutdownEvent(),
							m_stream->getEventTarget());
		EVENTQUEUE->removeHandler(ISocket::getDisconnectedEvent(),
							m_stream->getEventTarget());
		delete m_stream;
		m_stream = NULL;
	}
}

void
CClient::cleanupScreen()
{
	if (m_server != NULL) {
		if (m_ready) {
			m_screen->disable();
			m_ready = false;
		}
		EVENTQUEUE->removeHandler(IScreen::getShapeChangedEvent(),
							getEventTarget());
		delete m_server;
		m_server = NULL;
	}
}

void
CClient::cleanupTimer()
{
	if (m_timer != NULL) {
		EVENTQUEUE->removeHandler(CEvent::kTimer, m_timer);
		EVENTQUEUE->deleteTimer(m_timer);
		m_timer = NULL;
	}
}

void
CClient::handleConnected(const CEvent&, void*)
{
	LOG((CLOG_DEBUG1 "connected;  wait for hello"));
	cleanupConnecting();
	setupConnection();
}

void
CClient::handleConnectionFailed(const CEvent& event, void*)
{
	IDataSocket::CConnectionFailedInfo* info =
		reinterpret_cast<IDataSocket::CConnectionFailedInfo*>(event.getData());

	cleanupTimer();
	cleanupConnecting();
	delete m_stream;
	m_stream = NULL;
	LOG((CLOG_DEBUG1 "connection failed"));
	sendConnectionFailedEvent(info->m_what);
}

void
CClient::handleConnectTimeout(const CEvent&, void*)
{
	cleanupTimer();
	cleanupConnecting();
	cleanupConnection();
	delete m_stream;
	m_stream = NULL;
	LOG((CLOG_DEBUG1 "connection timed out - retrying"));
	sendConnectionFailedEvent("Timed out");
}

void
CClient::handleOutputError(const CEvent&, void*)
{
	cleanupTimer();
	cleanupScreen();
	cleanupConnection();
	LOG((CLOG_WARN "error sending to server"));
	sendEvent(getDisconnectedEvent(), NULL);
}

void
CClient::handleDisconnected(const CEvent&, void*)
{
	cleanupTimer();
	cleanupScreen();
	cleanupConnection();
	LOG((CLOG_DEBUG1 "disconnected"));
	sendEvent(getDisconnectedEvent(), NULL);
}

void
CClient::handleShapeChanged(const CEvent&, void*)
{
	LOG((CLOG_DEBUG "resolution changed"));
	m_server->onInfoChanged();
}

void
CClient::handleHello(const CEvent&, void*)
{
	SInt16 major, minor;
	if (!CProtocolUtil::readf(m_stream, kMsgHello, &major, &minor)) {
		sendConnectionFailedEvent("Protocol error from server");
		cleanupTimer();
		cleanupConnection();
		return;
	}

	// check versions
	LOG((CLOG_DEBUG1 "got hello version %d.%d", major, minor));
	if (major < kProtocolMajorVersion ||
		(major == kProtocolMajorVersion && minor < kProtocolMinorVersion)) {
		sendConnectionFailedEvent(XIncompatibleClient(major, minor).what());
		cleanupTimer();
		cleanupConnection();
		return;
	}

	// say hello back
	LOG((CLOG_DEBUG1 "say hello version %d.%d", kProtocolMajorVersion, kProtocolMinorVersion));
	CProtocolUtil::writef(m_stream, kMsgHelloBack,
							kProtocolMajorVersion,
							kProtocolMinorVersion, &m_name);

	// now connected but waiting to complete handshake
	setupScreen();
	cleanupTimer();

	// make sure we process any remaining messages later.  we won't
	// receive another event for already pending messages so we fake
	// one.
	if (m_stream->isReady()) {
		EVENTQUEUE->addEvent(CEvent(IStreamSYN::getInputReadyEvent(),
							m_stream->getEventTarget()));
	}
}

void
CClient::handleSuspend(const CEvent&, void*)
{
	LOG((CLOG_INFO "suspend"));
	m_suspended       = true;
	bool wasConnected = isConnected();
	disconnect(NULL);
	m_connectOnResume = wasConnected;
}

void
CClient::handleResume(const CEvent&, void*)
{
	LOG((CLOG_INFO "resume"));
	m_suspended = false;
	if (m_connectOnResume) {
		m_connectOnResume = false;
		connect();
	}
}

void
CClient::handleMotionEvent(const CEvent& event, void*)
{
	IPlatformScreen::CMotionInfo* info =
		reinterpret_cast<IPlatformScreen::CMotionInfo*>(event.getData());
	onMouseMove(info->m_x, info->m_y);
}

void
CClient::onMouseMove(SInt32 x, SInt32 y)
{
	LOG((CLOG_DEBUG2 "onMouseMove %d,%d", x, y));
	if(!m_active) {
		if(lockMouseTimer.getTime() >= 0.2) {
			lockMouseTimer.stop();
			CProtocolUtil::writef(m_stream,kMsgCInput);
		}
		else {
			LOG((CLOG_DEBUG2 "mouse motion event discarded"));
		}
	}
	// save position
	m_x       = x;
	m_y       = y;

	// get screen shape
	SInt32 ax, ay, aw, ah;
	getShape(ax, ay, aw, ah);
	SInt32 zoneSize = getJumpZoneSize();

	// see if we should change screens and move to edge if needed
	EDirection dir;
	if (x < ax + zoneSize) {
		x = ax;
		dir = kLeft;
		m_isInJumpzone = true;
	}
	else if (x >= ax + aw - zoneSize) {
		x = ax + aw - 1;
		dir = kRight;
		m_isInJumpzone = true;
	}
	else if (y < ay + zoneSize) {
		y = ay;
		dir = kTop;
		m_isInJumpzone = true;
	}
	else if (y >= ay + ah - zoneSize) {
		y = ay + ah - 1;
		dir = kBottom;
		m_isInJumpzone = true;
	}
	else {
		if(m_isTwoTapActive == 1) {
			if(m_isInJumpzone == true) {
				// have left jumpzone, send signal to server to arm two tap
				m_isInJumpzone = false;
				CProtocolUtil::writef(m_stream,kMsgCArmTwoTap);
			}
		}
		return;
	}
	// interpret dir for display to user and transmission to server
	CStringSYN dirName;
	SInt8 dirInt;
	switch(dir){
		case 1:
			dirName = "left";
			dirInt = 1;
			break;
		case 2:
			dirName = "right";
			dirInt = 2;
			break;
		case 3:
			dirName = "top";
			dirInt = 3;
			break;
		case 4:
			dirName = "bottom";
			dirInt = 4;
			break;
		default:
			dirName = "NULL";
			dirInt = 0;
			break;
	}

	if(isLockedToScreen()) {
		LOG((CLOG_DEBUG1 "locked to screen"));
		return;
	}

	LOG((CLOG_DEBUG1 "reached %s jump zone - (%d,%d)", dirName.c_str(), x, y));
	CProtocolUtil::writef(m_stream,kMsgDJumpZone, x, y, dirInt);	
}

bool
CClient::isLockedToScreen() const
{
	return m_screen->isLockedToScreen();
}

SInt32
CClient::getJumpZoneSize() const
{
	return this->m_screen->getJumpZoneSize();
}

