/*
 * SYNCHRONY - Modified 2006/09/15 by Maurice Blount
 * based on the program
 *
 * synergy -- mouse and keyboard sharing utility
 * Copyright (C) 2004 Chris Schoeneman
 * 
 * This package is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * found in the file COPYING that should have accompanied this file.
 * 
 * This package is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#ifndef CMSWINDOWSDESKS_H
#define CMSWINDOWSDESKS_H

#include "CSynchronyHook.h"
#include "MouseTypes.h"
#include "OptionTypes.h"
#include "CCondVar.h"
#include "CMutex.h"
#include "CString.h"
#include "stdmap.h"
#define WIN32_LEAN_AND_MEAN
#include <windows.h>

class CEvent;
class CEventQueueTimer;
class CThread;
class IJob;

//! Microsoft Windows desk handling
/*!
Desks in Microsoft Windows are only remotely like desktops on X11
systems.  A desk is another virtual surface for windows but desks
impose serious restrictions:  a thread can interact with only one
desk at a time, you can't switch desks if the thread has any hooks
installed or owns any windows, windows cannot exist on multiple
desks at once, etc.  Basically, they're useless except for running
the login window or the screensaver, which is what they're used
for.  Synergy must deal with them mainly because of the login
window and screensaver but users can create their own desks and
synergy should work on those too.

This class encapsulates all the desk nastiness.  Clients of this
object don't have to know anything about desks.
*/
class CMSWindowsDesks {
public:
	//! Constructor
	/*!
	\p isPrimary is true iff the desk is for a primary screen.
	\p hookLibrary must be a handle to the hook library.
	*/
	CMSWindowsDesks(bool isPrimary, HINSTANCE hookLibrary);
	~CMSWindowsDesks();

	//! @name manipulators
	//@{

	//! Enable desk tracking
	/*!
	Enables desk tracking.  While enabled, this object checks to see
	if the desk has changed and ensures that the hooks are installed
	on the new desk.  \c setShape should be called at least once
	before calling \c enable.
	*/
	void				enable();

	//! Disable desk tracking
	/*!
	Disables desk tracking.  \sa enable.
	*/
	void				disable();

	//! Notify of entering a desk
	/*!
	Prepares a desk for when the cursor enters it.
	*/
	void				enter();

	//! Notify of leaving a desk
	/*!
	Prepares a desk for when the cursor leaves it.
	*/
	void				leave();

	//! Notify of options changes
	/*!
	Resets all options to their default values.
	*/
	void				resetOptions();

	//! Notify of options changes
	/*!
	Set options to given values.  Ignores unknown options and doesn't
	modify options that aren't given in \c options.
	*/
	void				setOptions(const COptionsList& options);

	//! Tell desk about new size
	/*!
	This tells the desks that the display size has changed.
	*/
	void				setShape(SInt32 x, SInt32 y,
							SInt32 width, SInt32 height,
							SInt32 xCenter, SInt32 yCenter, bool isMultimon);

	//@}
	//! @name accessors
	//@{

	//! Get cursor position
	/*!
	Return the current position of the cursor in \c x and \c y.
	*/
	void				getCursorPos(SInt32& x, SInt32& y) const;

	//@}

private:
	class CDesk {
	public:
		CStringSYN		m_name;
		CThread*		m_thread;
		DWORD			m_threadID;
		DWORD			m_targetID;
		HDESK			m_desk;
		HWND			m_window;
		HWND			m_foregroundWindow;
		bool			m_lowLevel;
	};
	typedef std::map<CStringSYN, CDesk*> CDesks;

	// initialization and shutdown operations
	void				queryHookLibrary(HINSTANCE hookLibrary);
	HCURSOR				createBlankCursor() const;
	void				destroyCursor(HCURSOR cursor) const;
	ATOM				createDeskWindowClass() const;
	void				destroyClass(ATOM windowClass) const;
	HWND				createWindow(ATOM windowClass, const char* name) const;
	void				destroyWindow(HWND) const;

	// message handlers
	void				deskMouseMove(SInt32 x, SInt32 y) const;
	void				deskEnter(CDesk* desk);
	void				deskLeave(CDesk* desk);
	void				deskThread(void* vdesk);

	// desk switch checking and handling
	CDesk*				addDesk(const CStringSYN& name, HDESK hdesk);
	void				removeDesks();
	void				checkDesk();
	bool				isDeskAccessible(const CDesk* desk) const;
	void				handleCheckDesk(const CEvent& event, void*);

	// communication with desk threads
	void				waitForDesk() const;
	void				sendMessage(UINT, WPARAM, LPARAM) const;

	// work around for messed up keyboard events from low-level hooks
	HWND				getForegroundWindow() const;

	// desk API wrappers
	HDESK				openInputDesktop();
	void				closeDesktop(HDESK);
	CStringSYN			getDesktopName(HDESK);

	// our desk window proc
	static LRESULT CALLBACK primaryDeskProc(HWND, UINT, WPARAM, LPARAM);
	

private:
	// true if screen is being used as a primary screen, false otherwise
	bool				m_isPrimary;

	// true if windows 95/98/me
	bool				m_is95Family;

	// true if windows 98/2k or higher (i.e. not 95/nt)
	bool				m_isModernFamily;

	// true if mouse has entered the screen
	bool				m_isOnScreen;

	// our resources
	ATOM				m_deskClass;
	HCURSOR				m_cursor;

	// screen shape stuff
	SInt32				m_x, m_y;
	SInt32				m_w, m_h;
	SInt32				m_xCenter, m_yCenter;

	// true if system appears to have multiple monitors
	bool				m_multimon;

	// the timer used to check for desktop switching
	CEventQueueTimer*	m_timer;

	// the current desk and it's name
	CDesk*				m_activeDesk;
	CStringSYN			m_activeDeskName;

	// one desk per desktop and a cond var to communicate with it
	CMutex				m_mutex;
	CCondVar<bool>		m_deskReady;
	CDesks				m_desks;

	// hook library stuff
	InstallFunc			m_install;
	UninstallFunc		m_uninstall;
	
	// options
	bool				m_leaveForegroundOption;
};

#endif
