// ------------------------------------------------------------------------------------------------
// Adder Technology Ltd.
// Communications Port.
// (c) 2005 Adder Technology Ltd,Trafalgar Way, Bar Hill, Cambridge CB3 8SQ.
// Tel: +44 1954 780044 Fax: +44 1954 780081
// ------------------------------------------------------------------------------------------------
// RS232.cpp
// ------------------------------------------------------------------------------------------------
// Modified:	19.05.05
// By:			Pete Arnold
// Created:		03.03.05
// By:			Pete Arnold
// ------------------------------------------------------------------------------------------------
// History:		19.05.05 Replaced empirical delays with retried ReadFile calls.
//				11.05.05 Updated the send and receive buffer definitions to be DWORDs.
//				25.04.05 Removed local timeout code.
//				21.04.05 Increased timeouts to 500ms.
//				08.04.05 Minor modifications.
//				30.03.05 Review of functions and data.
//				03.03.05 Encapsulation of RS232 communications functions and data.
// ------------------------------------------------------------------------------------------------

#include "CMSWindowsRS232.h"
#include "CString.h"
#include <windows.h>

#define DEBUG_COMMS_OFF			// Only define DEBUG_COMMS if debug text is required on output.

// #ifdef wReserved				// The wReserved entry in the DCB is #defined to wEffects in richedit.h
// #undef wReserved				// so remove the definition in this file otherwise it won't compile.
// #endif

// ------------------------------------------------------------------------------------------------
// CMSWindowsRS232::RS232Initialise
// ------------------------------------------------------------------------------------------------
// Initialise the RS232 data structures.
// ------------------------------------------------------------------------------------------------
// IN:		comPort	:	byte indicating numerical value of the com port to use.
//			baudRate:	int indicating the baudrate to use.
// OUT:		None.
// Return:	void.
// ------------------------------------------------------------------------------------------------

void CRS232::RS232Initialise(BYTE comPort, int baudRate)
{
	// Initialise the Comms Status Block.
	RS232DCB.ComPortNumber = comPort;
	RS232DCB.Open = FALSE;
	RS232DCB.ComHandle = INVALID_HANDLE_VALUE;					// Set to invalid value to prevent misbehaviour.
	// Initialise the Port Device Control Block.
	RS232DCB.CommsDCB.DCBlength = sizeof(RS232DCB.CommsDCB);
	RS232DCB.CommsDCB.BaudRate = baudRate;						// Use decimal value of baud rate.
	RS232DCB.CommsDCB.ByteSize = 8;								// 7 or 8.
	RS232DCB.CommsDCB.EofChar = 0;
	RS232DCB.CommsDCB.ErrorChar = 0;
	RS232DCB.CommsDCB.EvtChar = 0;
	RS232DCB.CommsDCB.fAbortOnError = FALSE;					
	RS232DCB.CommsDCB.fBinary = TRUE;							// Must be TRUE, only binary data is supported.
	RS232DCB.CommsDCB.fDsrSensitivity = FALSE;
	RS232DCB.CommsDCB.fDtrControl = DTR_CONTROL_DISABLE;		// DTR_CONTROL_DISABLE etc. (see winbase.h).
	RS232DCB.CommsDCB.fDummy2 = FALSE;
	RS232DCB.CommsDCB.fErrorChar = FALSE;
	RS232DCB.CommsDCB.fNull = FALSE;
	RS232DCB.CommsDCB.fOutxCtsFlow = FALSE;						// For hardware control use: TRUE.
	RS232DCB.CommsDCB.fOutxDsrFlow = FALSE;
	RS232DCB.CommsDCB.fParity = FALSE;							// Whether or not parity is checked and errors reported.						
	RS232DCB.CommsDCB.fRtsControl = RTS_CONTROL_DISABLE;		// RTS_CONTROL_DISABLE etc. (see winbase.h).
																// For hardware control use: RTS_CONTROL_HANDSHAKE.
	RS232DCB.CommsDCB.fTXContinueOnXoff = FALSE;
	RS232DCB.CommsDCB.Parity = NOPARITY;						// EVENPARITY, ODDPARITY or NOPARITY.
	RS232DCB.CommsDCB.StopBits = ONESTOPBIT;					// ONESTOPBIT or TWOSTOPBITS.
	// RS232DCB.CommsDCB.fDummy2;
	// RS232DCB.CommsDCB.wReserved = 0;
	RS232DCB.CommsDCB.fInX = FALSE;								// For XON/XOFF use: TRUE
	RS232DCB.CommsDCB.fOutX = FALSE;							// TRUE
	RS232DCB.CommsDCB.XoffChar = 0;								// XON
	RS232DCB.CommsDCB.XoffLim = 0;								// 2048
	RS232DCB.CommsDCB.XonChar = 0;								// XOFF
	RS232DCB.CommsDCB.XonLim = 0;								// 2048
	// Initialise the communications timeouts.
	RS232DCB.CommsTimeouts.ReadIntervalTimeout = MAXDWORD;		// Maximum time between read chars. MAXDWORD.
	RS232DCB.CommsTimeouts.ReadTotalTimeoutMultiplier = 100;	// Multiplier of characters. 0.		1.
	RS232DCB.CommsTimeouts.ReadTotalTimeoutConstant = 500;		// Constant in milliseconds. 500.	2.
	RS232DCB.CommsTimeouts.WriteTotalTimeoutMultiplier = 100;	// Multiplier of characters. 10.	1.
	RS232DCB.CommsTimeouts.WriteTotalTimeoutConstant = 500;		// Constant in milliseconds. 500.	2.
}

// ------------------------------------------------------------------------------------------------
// CMSWindowsRS232::RS232Open
// ------------------------------------------------------------------------------------------------
// Open the RS232 port specified in the DCB.
// ------------------------------------------------------------------------------------------------
// IN/OUT:	Pointer to a DCB which specifies the parameters to use in the serial port.
// OUT:		None.
// Return:	RS232_RETURN_CODE	-	RS232_OK		:	Port opened OK.
//									RS232_CANT_OPEN	:	Port failed to open.
// ------------------------------------------------------------------------------------------------

CRS232::RS232_RETURN_CODE CRS232::RS232Open(void)
{
	RS232_RETURN_CODE	returnCode = RS232_OK;
	CStringSYN portID;
	
	if (RS232DCB.Open)
	{
		returnCode = RS232_ALREADY_OPEN;
	}
	else
	{
		// Set the port name directly for now.
		switch(RS232DCB.ComPortNumber) {
			case 1:
				portID = "COM1";
				break;
			case 2:
				portID = "COM2";
				break;
			case 3:
				portID = "COM3";
				break;
			case 4:
				portID = "COM4";
				break;
			case 5:
				portID = "COM5";
				break;
			case 6:
				portID = "COM6";
				break;
			case 7:
				portID = "COM7";
				break;
			case 8:
				portID = "COM8";
				break;
		}

		// Open the port without overlapping for synchronous tx/rx (for now).
		RS232DCB.ComHandle = CreateFile(portID.c_str(), GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_EXISTING, 0, NULL);
		if (RS232DCB.ComHandle == INVALID_HANDLE_VALUE)
		{
			returnCode = RS232_CANT_OPEN;
		}
		else 
		{
			// Set the comms state & timeouts to the settings defined above.
			if (!SetCommState(RS232DCB.ComHandle, &RS232DCB.CommsDCB))
			{
				
			}
			if (!SetCommTimeouts(RS232DCB.ComHandle, &RS232DCB.CommsTimeouts))
			{
				
			}

			// Mark the port as open.
			RS232DCB.Open = TRUE;
			// Reset the port.
			if (CRS232::RS232ResetSerial())
			{
				
			}
		}
		// Clear any outstanding errors.
		ClearCommError(RS232DCB.ComHandle, NULL, NULL);
	}

	return (returnCode);
}

// ------------------------------------------------------------------------------------------------
// CMSWindowsRS232::RS232Close
// ------------------------------------------------------------------------------------------------
// Close the RS232 port specified in the DCB.
// ------------------------------------------------------------------------------------------------
// IN:		None.
// OUT:		None.
// Return:	RS232_RETURN_CODE	-	RS232_OK		:	Port closed OK.
//									RS232_NOT_OPEN	:	Port not open.
// ------------------------------------------------------------------------------------------------

CRS232::RS232_RETURN_CODE CRS232::RS232Close(void)
{
	RS232_RETURN_CODE	returnCode = RS232_OK;

	

	if (RS232DCB.Open) 
	{
		// Mark port as closed.
		RS232DCB.Open = FALSE;
		// Close the serial port & release resources.
		if (RS232DCB.ComHandle != INVALID_HANDLE_VALUE )
		{
			if (!PurgeComm(RS232DCB.ComHandle, PURGE_TXABORT | PURGE_RXABORT | PURGE_TXCLEAR | PURGE_RXCLEAR))
			{
				
			}
			if (!EscapeCommFunction(RS232DCB.ComHandle, CLRDTR))
			{
				
			}
			if (!SetCommMask(RS232DCB.ComHandle, 0))
			{
				
			}
			if (!CloseHandle(RS232DCB.ComHandle))
			{
				
			}
		}
		// Clear handle value to avoid mistakes.
		RS232DCB.ComHandle = INVALID_HANDLE_VALUE;
	}
	else
	{
		returnCode = RS232_NOT_OPEN;
	}

	return (returnCode);
}

// ------------------------------------------------------------------------------------------------
// CMSWindowsRS232::RS232Send
// ------------------------------------------------------------------------------------------------
// Send data on the RS232 port specified in the DCB.
// ------------------------------------------------------------------------------------------------
// IN/OUT:		BYTE*	sendBuffer		:	the address of the buffer containing the data to send.
//				DWORD*	sendBufferSize	:	pointer to the amount of data to send and into which
//											the actual number of bytes sent is placed.
// Return:	RS232_RETURN_CODE	-	RS232_OK		:	Data sent OK.
//									RS232_NOT_OPEN	:	Port not open.
//									RS232_NO_DATA	:	No data to send.
//									RS232_FAILED	:	The send command failed.
// ------------------------------------------------------------------------------------------------

CRS232::RS232_RETURN_CODE CRS232::RS232Send(BYTE* sendBuffer, DWORD* sendBufferSize)
{
	RS232_RETURN_CODE	returnCode = RS232_OK;
	DWORD				count_sent = 0;

	

	if (RS232DCB.Open)
	{

		if (*sendBufferSize > 0)
		{

			// Clear any outstanding errors. 
			ClearCommError(RS232DCB.ComHandle, NULL, NULL);
			
			// Write the data to the RS232 port.
			// The overlapped pointer is set to NULL for synchronous operation.
			if (!WriteFile(RS232DCB.ComHandle, sendBuffer, *sendBufferSize, &count_sent, NULL))
			{
				returnCode = RS232_FAILED;
			}

			// Return the number of bytes actually sent to the calling function.
			*sendBufferSize = count_sent;
			// Clear any outstanding errors.
			ClearCommError(RS232DCB.ComHandle, NULL, NULL);
		}
		else
		{
			returnCode = RS232_NO_DATA;
		}

	}
	else
	{
		returnCode = RS232_NOT_OPEN;
		*sendBufferSize = 0;  
	}

	return (returnCode);
}

// ------------------------------------------------------------------------------------------------
// CMSWindowsRS232::RS232ResetSerial
// ------------------------------------------------------------------------------------------------
// Reset the RS232 port specified in the DCB.
// ------------------------------------------------------------------------------------------------
// IN:		None.
// OUT:		none.
// Return:	RS232_RETURN_CODE	-	RS232_OK			:	RS232 reset OK.
//									RS232_NOT_OPEN		:	Port not open.
//									RS232_FAILED		:	The reset command failed.
// ------------------------------------------------------------------------------------------------

CRS232::RS232_RETURN_CODE CRS232::RS232ResetSerial(void)
{
	RS232_RETURN_CODE	returnCode = RS232_OK;
	DWORD				errors;

	
	if (RS232DCB.Open) 
	{
		// Clear any outstanding errors (check for error types with the ERROR variable, ignore the STATE variable).
		ClearCommError(RS232DCB.ComHandle, &errors, NULL);
		// Report any error data as debug. Ignore the state structure.
		if (errors != 0)
		{
			returnCode = RS232_FAILED;
		}
		// Set-up the communications port with appropriately sized input and output buffers.
		if (!SetupComm(RS232DCB.ComHandle, RS232_IN_QUEUE_SIZE, RS232_OUT_QUEUE_SIZE))
		{
			returnCode = RS232_FAILED;
		}
		// Terminate any outstanding TX/RX operations and flush the input and output buffers.
		if (!PurgeComm(RS232DCB.ComHandle, PURGE_TXABORT | PURGE_RXABORT | PURGE_TXCLEAR | PURGE_RXCLEAR))
		{
			returnCode = RS232_FAILED;
		}
		// Reset the comms state & timeouts to the settings defined above.
		if (!SetCommState(RS232DCB.ComHandle, &RS232DCB.CommsDCB))
		{
			returnCode = RS232_FAILED;
		}
		if (!SetCommTimeouts(RS232DCB.ComHandle, &RS232DCB.CommsTimeouts))
		{
			returnCode = RS232_FAILED;
		}
		// Clear the masks - no events are handled at present.
		SetCommMask(RS232DCB.ComHandle, 0);
	}
	else
	{
		returnCode = RS232_NOT_OPEN;
	}
    
	return (returnCode);

}

// ------------------------------------------------------------------------------------------------
// UploadCheckCommsPort
// ------------------------------------------------------------------------------------------------
// Check the availability of comms at a specified port.
// Used to check which ports are present so that only these can be presented to the user for
// selection.
// ------------------------------------------------------------------------------------------------
// IN:		int	i					:	Comms port to check.
// OUT:		None.
// Return:	UPLOADER_RETURN_CODE	:	UPLOADER_OK				Comms OK at port i.
//										UPLOADER_COMMS_FAILED	Comms not available at port i.
// ------------------------------------------------------------------------------------------------

CRS232::RS232_RETURN_CODE CRS232::UploadCheckCommsPort(int i)
{
	RS232_RETURN_CODE returnCode;

	// Select and initialise the required COM port, if it opens, respond with OK otherwise with
	// an error. Close the COM port.
	RS232Initialise(i, BAUDRATE);
	if (RS232Open() == CRS232::RS232_OK)
	{
		returnCode = RS232_OK;
	}
	else
	{
		returnCode = RS232_FAILED;
	}
	RS232Close();

	return (returnCode);
}


