/*
 * SYNCHRONY - Modified 2006/09/15 by Maurice Blount
 * based on the program
 *
 * synergy -- mouse and keyboard sharing utility
 * Copyright (C) 2002 Chris Schoeneman
 * 
 * This package is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * found in the file COPYING that should have accompanied this file.
 * 
 * This package is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#include "CSynchronyHook.h"
#include "ProtocolTypes.h"
#include <zmouse.h>

//
// debugging compile flag.  when not zero the server doesn't grab
// the keyboard when the mouse leaves the server screen.  this
// makes it possible to use the debugger (via the keyboard) when
// all user input would normally be caught by the hook procedures.
//
#define NO_GRAB_KEYBOARD 0

//
// debugging compile flag.  when not zero the server will not
// install low level hooks.
//
#define NO_LOWLEVEL_HOOKS 0

// declare extended mouse hook struct.  useable on win2k
typedef struct tagMOUSEHOOKSTRUCTWin2000 {
	MOUSEHOOKSTRUCT mhs;
	DWORD mouseData;
} MOUSEHOOKSTRUCTWin2000;

#if !defined(SM_MOUSEWHEELPRESENT)
#define SM_MOUSEWHEELPRESENT 75
#endif

// X button stuff
#if !defined(WM_XBUTTONDOWN)
#define WM_XBUTTONDOWN		0x020B
#define WM_XBUTTONUP		0x020C
#define WM_XBUTTONDBLCLK	0x020D
#define WM_NCXBUTTONDOWN	0x00AB
#define WM_NCXBUTTONUP		0x00AC
#define WM_NCXBUTTONDBLCLK	0x00AD
#define MOUSEEVENTF_XDOWN	0x0080
#define MOUSEEVENTF_XUP		0x0100
#define XBUTTON1			0x0001
#define XBUTTON2			0x0002
#endif


//
// globals
//

#if defined(_MSC_VER)
#pragma comment(linker, "-section:shared,rws")
#pragma data_seg("shared")
#endif
// all data in this shared section *must* be initialized

static HINSTANCE		g_hinstance       = NULL;
static DWORD			g_processID       = 0;
static DWORD			g_threadID        = 0;
static HHOOK			g_mouse           = NULL;
static HHOOK			g_getMessage      = NULL;
static HHOOK			g_mouseLL         = NULL;
static EHookMode		g_mode            = kHOOK_DISABLE;
static UInt32			g_zoneSides       = 0;
static SInt32			g_zoneSize        = 0;
static SInt32			g_xScreen         = 0;
static SInt32			g_yScreen         = 0;
static SInt32			g_wScreen         = 0;
static SInt32			g_hScreen         = 0;
static WPARAM			g_deadVirtKey     = 0;
static LPARAM			g_deadLParam      = 0;
static BYTE				g_deadKeyState[256] = { 0 };
static DWORD			g_hookThread      = 0;
static DWORD			g_attachedThread  = 0;

#if defined(_MSC_VER)
#pragma data_seg()
#endif

// keep linker quiet about floating point stuff.  we don't use any
// floating point operations but our includes may define some
// (unused) floating point values.
#ifndef _DEBUG
extern "C" {
int _fltused=0;
}
#endif


//
// internal functions
//

static
void
detachThread()
{
	if (g_attachedThread != 0 && g_hookThread != g_attachedThread) {
		AttachThreadInput(g_hookThread, g_attachedThread, FALSE);
		g_attachedThread = 0;
	}
}


static
bool
doMouseHookHandler(WPARAM wParam, SInt32 x, SInt32 y, SInt32 data)
{
	switch (wParam) {
	case WM_LBUTTONDOWN:
	case WM_MBUTTONDOWN:
	case WM_RBUTTONDOWN:
	case WM_XBUTTONDOWN:
	case WM_LBUTTONDBLCLK:
	case WM_MBUTTONDBLCLK:
	case WM_RBUTTONDBLCLK:
	case WM_XBUTTONDBLCLK:
	case WM_LBUTTONUP:
	case WM_MBUTTONUP:
	case WM_RBUTTONUP:
	case WM_XBUTTONUP:
	case WM_NCLBUTTONDOWN:
	case WM_NCMBUTTONDOWN:
	case WM_NCRBUTTONDOWN:
	case WM_NCXBUTTONDOWN:
	case WM_NCLBUTTONDBLCLK:
	case WM_NCMBUTTONDBLCLK:
	case WM_NCRBUTTONDBLCLK:
	case WM_NCXBUTTONDBLCLK:
	case WM_NCLBUTTONUP:
	case WM_NCMBUTTONUP:
	case WM_NCRBUTTONUP:
	case WM_NCXBUTTONUP:
		// always relay the event.  eat it if relaying.
		PostThreadMessage(g_threadID, SYNERGY_MSG_MOUSE_BUTTON, wParam, data);
		return (g_mode == kHOOK_RELAY_EVENTS);

	case WM_MOUSEWHEEL:
		if (g_mode == kHOOK_RELAY_EVENTS) {
			// relay event
			PostThreadMessage(g_threadID, SYNERGY_MSG_MOUSE_WHEEL, data, 0);
		}
		return (g_mode == kHOOK_RELAY_EVENTS);

	case WM_NCMOUSEMOVE:
	case WM_MOUSEMOVE:
		if (g_mode == kHOOK_RELAY_EVENTS) {
			// relay and eat event
			PostThreadMessage(g_threadID, SYNERGY_MSG_MOUSE_MOVE, x, y);
			return true;
		}
		else if (g_mode == kHOOK_WATCH_JUMP_ZONE) {
			// low level hooks can report bogus mouse positions that are
			// outside of the screen.  jeez.  naturally we end up getting
			// fake motion in the other direction to get the position back
			// on the screen, which plays havoc with switch on double tap.
			// CServer deals with that.  we'll clamp positions onto the
			// screen.  also, if we discard events for positions outside
			// of the screen then the mouse appears to get a bit jerky
			// near the edge.  we can either accept that or pass the bogus
			// events.  we'll try passing the events.
			bool bogus = false;
			if (x < g_xScreen) {
				x     = g_xScreen;
				bogus = true;
			}
			else if (x >= g_xScreen + g_wScreen) {
				x     = g_xScreen + g_wScreen - 1;
				bogus = true;
			}
			if (y < g_yScreen) {
				y     = g_yScreen;
				bogus = true;
			}
			else if (y >= g_yScreen + g_hScreen) {
				y     = g_yScreen + g_hScreen - 1;
				bogus = true;
			}

			// check for mouse inside jump zone
			bool inside = false;
			if (!inside && (g_zoneSides & kLeftMask) != 0) {
				inside = (x < g_xScreen + g_zoneSize);
			}
			if (!inside && (g_zoneSides & kRightMask) != 0) {
				inside = (x >= g_xScreen + g_wScreen - g_zoneSize);
			}
			if (!inside && (g_zoneSides & kTopMask) != 0) {
				inside = (y < g_yScreen + g_zoneSize);
			}
			if (!inside && (g_zoneSides & kBottomMask) != 0) {
				inside = (y >= g_yScreen + g_hScreen - g_zoneSize);
			}

			// relay the event
			PostThreadMessage(g_threadID, SYNERGY_MSG_MOUSE_MOVE, x, y);

			// if inside and not bogus then eat the event
			return inside && !bogus;
		}
	}

	// pass the event
	return false;
}

static
bool
mouseHookHandler(WPARAM wParam, SInt32 x, SInt32 y, SInt32 data)
{
	return doMouseHookHandler(wParam, x, y, data);
}

static
LRESULT CALLBACK
mouseHook(int code, WPARAM wParam, LPARAM lParam)
{
	if (code >= 0) {
		// decode message
		const MOUSEHOOKSTRUCT* info = (const MOUSEHOOKSTRUCT*)lParam;
		SInt32 x = (SInt32)info->pt.x;
		SInt32 y = (SInt32)info->pt.y;
		SInt32 w = 0;
		
		// handle the message.  note that we don't handle X buttons
		// here.  that's okay because they're only supported on
		// win2k and winxp and up and on those platforms we'll get
		// get the mouse events through the low level hook.
		if (mouseHookHandler(wParam, x, y, w)) {
			return 1;
		}
	}

	return CallNextHookEx(g_mouse, code, wParam, lParam);
}


#if (_WIN32_WINNT >= 0x0400) && defined(_MSC_VER) && !NO_LOWLEVEL_HOOKS

//
// low-level mouse hook -- this allows us to capture and handle mouse
// events very early.  the earlier the better.
//

static
LRESULT CALLBACK
mouseLLHook(int code, WPARAM wParam, LPARAM lParam)
{
	if (code >= 0) {
		// decode the message
		MSLLHOOKSTRUCT* info = reinterpret_cast<MSLLHOOKSTRUCT*>(lParam);
		SInt32 x = static_cast<SInt32>(info->pt.x);
		SInt32 y = static_cast<SInt32>(info->pt.y);
		SInt32 w = static_cast<SInt16>(HIWORD(info->mouseData));

		// handle the message
		if (mouseHookHandler(wParam, x, y, w)) {
			return 1;
		}
	}

	return CallNextHookEx(g_mouseLL, code, wParam, lParam);
}

#endif




//
// external functions
//

BOOL WINAPI
DllMain(HINSTANCE instance, DWORD reason, LPVOID)
{
	if (reason == DLL_PROCESS_ATTACH) {
		DisableThreadLibraryCalls(instance);
		if (g_processID == 0) {
			g_hinstance = instance;
			g_processID = GetCurrentProcessId();
		}
	}
	else if (reason == DLL_PROCESS_DETACH) {
		if (g_processID == GetCurrentProcessId()) {
			uninstall();
			g_processID = 0;
			g_hinstance = NULL;
		}
	}
	return TRUE;
}

extern "C" {

int
init(DWORD threadID)
{
	assert(g_hinstance != NULL);

	// try to open process that last called init() to see if it's
	// still running or if it died without cleaning up.
	if (g_processID != 0 && g_processID != GetCurrentProcessId()) {
		HANDLE process = OpenProcess(STANDARD_RIGHTS_REQUIRED,
								FALSE, g_processID);
		if (process != NULL) {
			// old process (probably) still exists so refuse to
			// reinitialize this DLL (and thus steal it from the
			// old process).
			CloseHandle(process);
			return 0;
		}

		// clean up after old process.  the system should've already
		// removed the hooks so we just need to reset our state.
		g_hinstance       = GetModuleHandle("synchrnyhk");
		g_processID       = GetCurrentProcessId();
		g_threadID        = 0;
		g_mouse           = NULL;
		g_getMessage      = NULL;
		g_mouseLL         = NULL;
	}

	// save thread id.  we'll post messages to this thread's
	// message queue.
	g_threadID  = threadID;

	// set defaults
	g_mode      = kHOOK_DISABLE;
	g_zoneSides = 0;
	g_zoneSize  = 0;
	g_xScreen   = 0;
	g_yScreen   = 0;
	g_wScreen   = 0;
	g_hScreen   = 0;

	return 1;
}

int
cleanup(void)
{
	assert(g_hinstance != NULL);

	if (g_processID == GetCurrentProcessId()) {
		g_threadID = 0;
	}

	return 1;
}

EHookResult
install()
{
	assert(g_hinstance  != NULL);
//	assert(g_keyboard   == NULL);
	assert(g_mouse      == NULL);
	assert(g_getMessage == NULL);

	// must be initialized
	if (g_threadID == 0) {
		return kHOOK_FAILED;
	}

	// discard old dead keys
	g_deadVirtKey = 0;
	g_deadLParam  = 0;

	
	// install low-level hooks.  we require that they both get installed.
#if (_WIN32_WINNT >= 0x0400) && defined(_MSC_VER) && !NO_LOWLEVEL_HOOKS
	g_mouseLL = SetWindowsHookEx(WH_MOUSE_LL,
								&mouseLLHook,
								g_hinstance,
								0);
#endif

	// install regular hooks
	if (g_mouseLL == NULL) {
		g_mouse = SetWindowsHookEx(WH_MOUSE,
								&mouseHook,
								g_hinstance,
								0);
	}

	// check that we got all the hooks we wanted
	if (g_mouseLL    == NULL && g_mouse        == NULL) {
		uninstall();
		return kHOOK_FAILED;
	}

	if (g_mouseLL != NULL) {
		g_hookThread = GetCurrentThreadId();
		return kHOOK_OKAY_LL;
	}

	return kHOOK_OKAY;
}

int
uninstall(void)
{
	assert(g_hinstance != NULL);

	// discard old dead keys
	g_deadVirtKey = 0;
	g_deadLParam  = 0;

	// detach from thread
	detachThread();

	// uninstall hooks
	if (g_mouseLL != NULL) {
		UnhookWindowsHookEx(g_mouseLL);
		g_mouseLL = NULL;
	}
	if (g_mouse != NULL) {
		UnhookWindowsHookEx(g_mouse);
		g_mouse = NULL;
	}
	if (g_getMessage != NULL) {
		UnhookWindowsHookEx(g_getMessage);
		g_getMessage = NULL;
	}

	return 1;
}

void
setSides(UInt32 sides)
{
	g_zoneSides = sides;
}

void
setZone(SInt32 x, SInt32 y, SInt32 w, SInt32 h, SInt32 jumpZoneSize)
{
	g_zoneSize = jumpZoneSize;
	g_xScreen  = x;
	g_yScreen  = y;
	g_wScreen  = w;
	g_hScreen  = h;
}

void
setMode(EHookMode mode)
{
	if (mode == g_mode) {
		// no change
		return;
	}
	g_mode = mode;
}

}
