/*
 * SYNCHRONY - Modified 2006/09/15 by Maurice Blount
 * based on the program
 *
 * synergy -- mouse and keyboard sharing utility
 * Copyright (C) 2002 Chris Schoeneman
 * 
 * This package is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * found in the file COPYING that should have accompanied this file.
 * 
 * This package is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#include "CClientProxy1_0.h"
#include "CServer.h"
#include "CPrimaryClient.h"
#include "CProtocolUtil.h"
#include "XSynergy.h"
#include "IStream.h"
#include "CLog.h"
#include "IEventQueue.h"
#include "TMethodEventJob.h"
#include <cstring>

//
// CClientProxy1_0
//

CClientProxy1_0::CClientProxy1_0(const CStringSYN& name, IStreamSYN* stream) :
	CClientProxy(name, stream),
	m_heartbeatTimer(NULL),
	m_parser(&CClientProxy1_0::parseHandshakeMessage)
{
	// install event handlers
	EVENTQUEUE->adoptHandler(IStreamSYN::getInputReadyEvent(),
							stream->getEventTarget(),
							new TMethodEventJob<CClientProxy1_0>(this,
								&CClientProxy1_0::handleData, NULL));
	EVENTQUEUE->adoptHandler(IStreamSYN::getOutputErrorEvent(),
							stream->getEventTarget(),
							new TMethodEventJob<CClientProxy1_0>(this,
								&CClientProxy1_0::handleWriteError, NULL));
	EVENTQUEUE->adoptHandler(IStreamSYN::getInputShutdownEvent(),
							stream->getEventTarget(),
							new TMethodEventJob<CClientProxy1_0>(this,
								&CClientProxy1_0::handleDisconnect, NULL));
	EVENTQUEUE->adoptHandler(IStreamSYN::getOutputShutdownEvent(),
							stream->getEventTarget(),
							new TMethodEventJob<CClientProxy1_0>(this,
								&CClientProxy1_0::handleWriteError, NULL));
	EVENTQUEUE->adoptHandler(CEvent::kTimer, this,
							new TMethodEventJob<CClientProxy1_0>(this,
								&CClientProxy1_0::handleFlatline, NULL));

	setHeartbeatRate(kHeartRate, kHeartRate * kHeartBeatsUntilDeath);

	LOG((CLOG_DEBUG1 "querying client \"%s\" info", getName().c_str()));
	extern CServer* s_server;
	CProtocolUtil::writef(getStream(), kMsgQInfo, s_server->isTwoTapActive());
	m_hasLeft = true;
}

CClientProxy1_0::~CClientProxy1_0()
{
	removeHandlers();
}

void
CClientProxy1_0::disconnect()
{
	removeHandlers();
	getStream()->close();
	EVENTQUEUE->addEvent(CEvent(getDisconnectedEvent(), getEventTarget()));
}

void
CClientProxy1_0::removeHandlers()
{
	// uninstall event handlers
	EVENTQUEUE->removeHandler(IStreamSYN::getInputReadyEvent(),
							getStream()->getEventTarget());
	EVENTQUEUE->removeHandler(IStreamSYN::getOutputErrorEvent(),
							getStream()->getEventTarget());
	EVENTQUEUE->removeHandler(IStreamSYN::getInputShutdownEvent(),
							getStream()->getEventTarget());
	EVENTQUEUE->removeHandler(IStreamSYN::getOutputShutdownEvent(),
							getStream()->getEventTarget());
	EVENTQUEUE->removeHandler(CEvent::kTimer, this);

	// remove timer
	removeHeartbeatTimer();
}

void
CClientProxy1_0::addHeartbeatTimer()
{
	if (m_heartbeatAlarm > 0.0) {
		m_heartbeatTimer = EVENTQUEUE->newOneShotTimer(m_heartbeatAlarm, this);
	}
}

void
CClientProxy1_0::removeHeartbeatTimer()
{
	if (m_heartbeatTimer != NULL) {
		EVENTQUEUE->deleteTimer(m_heartbeatTimer);
		m_heartbeatTimer = NULL;
	}
}

void
CClientProxy1_0::resetHeartbeatTimer()
{
	// reset the alarm
	removeHeartbeatTimer();
	addHeartbeatTimer();
}

void
CClientProxy1_0::resetHeartbeatRate()
{
	setHeartbeatRate(kHeartRate, kHeartRate * kHeartBeatsUntilDeath);
}

void
CClientProxy1_0::setHeartbeatRate(double, double alarm)
{
	m_heartbeatAlarm = alarm;
}

void
CClientProxy1_0::handleData(const CEvent&, void*)
{
	// handle messages until there are no more.  first read message code.
	UInt8 code[4];
	UInt32 n = getStream()->read(code, 4);
	while (n != 0) {
		// verify we got an entire code
		if (n != 4) {
			LOG((CLOG_ERR "incomplete message from \"%s\": %d bytes", getName().c_str(), n));
			disconnect();
			return;
		}

		// parse message
		
		LOG((CLOG_DEBUG2 "msg from \"%s\": %c%c%c%c", getName().c_str(), code[0], code[1], code[2], code[3]));
		
		if (memcmp(code, kMsgCInput, 4) == 0) {
			extern CServer* s_server;
			if(s_server->m_active != this && this->m_hasLeft) {
				LOG((CLOG_DEBUG "input detected on \"%s\" - redefining active screen", getName().c_str()));
				// leave what was the active screen
				s_server->m_active->leave();
								
				SInt32 x, y, dx, dy, dw, dh;
				// finding center of screen
				this->getShape(dx, dy, dw, dh);
				x = dx + dw/2;
				y = dy + dh/2;
				// cut over
				this->enter(x, y);
				this->m_hasLeft = false;
				s_server->m_active = this;
			}
		}
		else if (memcmp(code, kMsgCHasLeft,4) == 0) {
			LOG((CLOG_DEBUG "\"%s\" has acknowledged leave signal", getName().c_str()));
			this->m_hasLeft = true;
		}
		else if (memcmp(code, kMsgCArmTwoTap, 4) == 0) {
			extern CServer* s_server;
			s_server->m_switchTwoTapArmed = true;
		}
		else if (memcmp(code, kMsgDJumpZone, 4) == 0) {
			SInt32 x, y;
			SInt8 dirInt;
			CProtocolUtil::readf(getStream(),kMsgDJumpZone + 4, &x, &y, &dirInt);
			LOG((CLOG_DEBUG "recieved DJMP from %s at (%d,%d)", getName().c_str(), x, y));
			checkJumpZone(getName(), x, y, dirInt);
		}
		else if (!(this->*m_parser)(code)) {
			LOG((CLOG_ERR "invalid message from client \"%s\": %c%c%c%c", getName().c_str(), code[0], code[1], code[2], code[3]));
			disconnect();
			return;
		}

		// next message
		n = getStream()->read(code, 4);
	}

	// restart heartbeat timer
	resetHeartbeatTimer();
}

bool
CClientProxy1_0::parseHandshakeMessage(const UInt8* code)
{
	if (memcmp(code, kMsgCNoop, 4) == 0) {
		// discard no-ops
		LOG((CLOG_DEBUG2 "no-op from", getName().c_str()));
		return true;
	}
	else if (memcmp(code, kMsgDInfo, 4) == 0) {
		// future messages get parsed by parseMessage
		m_parser = &CClientProxy1_0::parseMessage;
		if (recvInfo()) {
			EVENTQUEUE->addEvent(CEvent(getReadyEvent(), getEventTarget()));
			addHeartbeatTimer();
			return true;
		}
	}
	return false;
}

bool
CClientProxy1_0::parseMessage(const UInt8* code)
{
	if (memcmp(code, kMsgDInfo, 4) == 0) {
		if (recvInfo()) {
			EVENTQUEUE->addEvent(
							CEvent(getShapeChangedEvent(), getEventTarget()));
			return true;
		}
		return false;
	}
	else if (memcmp(code, kMsgCNoop, 4) == 0) {
		// discard no-ops
		LOG((CLOG_DEBUG2 "no-op from", getName().c_str()));
		return true;
	}
	return false;
}

void
CClientProxy1_0::handleDisconnect(const CEvent&, void*)
{
	LOG((CLOG_NOTE "client \"%s\" has disconnected", getName().c_str()));
	disconnect();
}

void
CClientProxy1_0::handleWriteError(const CEvent&, void*)
{
	LOG((CLOG_WARN "error writing to client \"%s\"", getName().c_str()));
	disconnect();
}

void
CClientProxy1_0::handleFlatline(const CEvent&, void*)
{
	// didn't get a heartbeat fast enough.  assume client is dead.
	LOG((CLOG_NOTE "client \"%s\" is dead", getName().c_str()));
	disconnect();
}

void
CClientProxy1_0::getShape(SInt32& x, SInt32& y, SInt32& w, SInt32& h) const
{
	x = m_info.m_x;
	y = m_info.m_y;
	w = m_info.m_w;
	h = m_info.m_h;
}

void
CClientProxy1_0::getCursorPos(SInt32& x, SInt32& y) const
{
	// note -- this returns the cursor pos from when we last got client info
	x = m_info.m_mx;
	y = m_info.m_my;
}

void
CClientProxy1_0::enter(SInt32 xAbs, SInt32 yAbs)
{
	LOG((CLOG_DEBUG1 "send enter to \"%s\", %d,%d ", getName().c_str(), xAbs, yAbs));
	CProtocolUtil::writef(getStream(), kMsgCEnter,
								xAbs, yAbs);
}

bool
CClientProxy1_0::leave()
{
	LOG((CLOG_DEBUG1 "send leave to \"%s\"", getName().c_str()));
	CProtocolUtil::writef(getStream(), kMsgCLeave);

	// we can never prevent the user from leaving
	return true;
}

void
CClientProxy1_0::resetOptions()
{
	LOG((CLOG_DEBUG1 "send reset options to \"%s\"", getName().c_str()));
	CProtocolUtil::writef(getStream(), kMsgCResetOptions);

	// reset heart rate and death
	resetHeartbeatRate();
	removeHeartbeatTimer();
	addHeartbeatTimer();
}

void
CClientProxy1_0::setOptions(const COptionsList& options)
{
	LOG((CLOG_DEBUG1 "send set options to \"%s\" size=%d", getName().c_str(), options.size()));
	CProtocolUtil::writef(getStream(), kMsgDSetOptions, &options);

	// check options
	for (UInt32 i = 0, n = options.size(); i < n; i += 2) {
		if (options[i] == kOptionHeartbeat) {
			double rate = 1.0e-3 * static_cast<double>(options[i + 1]);
			if (rate <= 0.0) {
				rate = -1.0;
			}
			setHeartbeatRate(rate, rate * kHeartBeatsUntilDeath);
			removeHeartbeatTimer();
			addHeartbeatTimer();
		}
	}
}


void 
CClientProxy1_0::checkJumpZone(CStringSYN srcName, SInt32 x, SInt32 y, SInt8 dirInt)
{
			//FIXME - may be a neater way to do this - we need to call member functions of CServer to switch screens
			//		  and so I have used the s_server which is instantiated in synchronys.
			extern CServer* s_server;

			if (s_server->lockMouseTimer <= 0.2) {
				LOG((CLOG_DEBUG2 "rogue mouse motion ignored"));
				return;
			}	
			EDirection dir;
			switch(dirInt){
				case 1:
					dir = kLeft;
					break;
				case 2:
					dir = kRight;
					break;
				case 3:
					dir = kTop;
					break;
				case 4:
					dir = kBottom;
					break;
				default:
					dir = kNoDirection;
					break;
			}
			CBaseClientProxy* neighbour = s_server->mapToNeighbor(this, dir, x, y);
			if(s_server->isSwitchOkay(neighbour, dir, x, y)) {
				s_server->switchScreen(neighbour, x, y);
			}
		return;
}

bool
CClientProxy1_0::recvInfo()
{
	// parse the message
	SInt16 x, y, w, h, dummy1, mx, my;
	if (!CProtocolUtil::readf(getStream(), kMsgDInfo + 4,
							&x, &y, &w, &h, &dummy1, &mx, &my)) {
		return false;
	}
	LOG((CLOG_DEBUG "received client \"%s\" info shape=%d,%d %dx%d", getName().c_str(), x, y, w, h));

	// validate
	if (w <= 0 || h <= 0) {
		return false;
	}

	// save
	m_info.m_x  = x;
	m_info.m_y  = y;
	m_info.m_w  = w;
	m_info.m_h  = h;
	m_info.m_mx = mx;
	m_info.m_my = my;

	// acknowledge receipt
	LOG((CLOG_DEBUG1 "send info ack to \"%s\"", getName().c_str()));
	CProtocolUtil::writef(getStream(), kMsgCInfoAck);
	return true;
}
