/*
 * SYNCHRONY - Modified 2006/09/15 by Maurice Blount
 * based on the program
 *
 * synergy -- mouse and keyboard sharing utility
 * Copyright (C) 2002 Chris Schoeneman
 * 
 * This package is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * found in the file COPYING that should have accompanied this file.
 * 
 * This package is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#include "CServer.h"
#include "CClientProxy.h"
#include "CClientProxyUnknown.h"
#include "CPrimaryClient.h"
#include "IPlatformScreen.h"
#include "OptionTypes.h"
#include "ProtocolTypes.h"
#include "XScreen.h"
#include "XSynergy.h"
#include "IDataSocket.h"
#include "IListenSocket.h"
#include "XSocket.h"
#include "IEventQueue.h"
#include "CLog.h"
#include "CStopwatch.h"
#include "TMethodEventJob.h"
#include "CArch.h"
#include "CRS232.h"
#include <string.h>

//
// CServer
//

CEvent::Type			CServer::s_errorEvent         = CEvent::kUnknown;
CEvent::Type			CServer::s_connectedEvent     = CEvent::kUnknown;
CEvent::Type			CServer::s_disconnectedEvent  = CEvent::kUnknown;

CServer::CServer(const CConfig& config, CPrimaryClient* primaryClient) :
	m_primaryClient(primaryClient),
	m_active(primaryClient),
	m_xDelta(0),
	m_yDelta(0),
	m_xDelta2(0),
	m_yDelta2(0),
	m_config(),
	m_switchDir(kNoDirection),
	m_switchScreen(NULL),
	m_switchWaitDelay(0.0),
	m_switchWaitTimer(NULL),
	m_switchTwoTapDelay(0.0),
	m_switchTwoTapEngaged(false),
	m_switchTwoTapArmed(false),
	m_switchTwoTapZone(3),
	m_isRS232Initialised(false)
{
	// must have a primary client and it must have a canonical name
	assert(m_primaryClient != NULL);
	assert(config.isScreen(primaryClient->getName()));

	CStringSYN primaryName = getName(primaryClient);

	// install event handlers
	EVENTQUEUE->adoptHandler(CEvent::kTimer, this,
							new TMethodEventJob<CServer>(this,
								&CServer::handleSwitchWaitTimeout));
	EVENTQUEUE->adoptHandler(IPlatformScreen::getMotionEvent(),
							m_primaryClient->getEventTarget(),
							new TMethodEventJob<CServer>(this,
								&CServer::handleMotionEvent));

	
	// add connection
	addClient(m_primaryClient);
	// set initial configuration
	setConfig(config);
	// enable primary client
	m_primaryClient->enable();
	// initialise RS232
#if SYSAPI_UNIX
	uploaderComPort.RS232Initialise(*(m_config.m_serialPort.c_str()), BAUDRATE);
#elif SYSAPI_MSWINDOWS
	uploaderComPort.RS232Initialise(m_config.getSerialPort(), BAUDRATE);
#endif
	m_isRS232Initialised = true;
}

CServer::~CServer()
{
	EVENTQUEUE->removeHandler(IPlatformScreen::getMotionEvent(),
							m_primaryClient->getEventTarget());
	EVENTQUEUE->removeHandler(CEvent::kTimer, this);
	stopSwitch();

	// force immediate disconnection of secondary clients
	disconnect();
	for (COldClients::iterator index = m_oldClients.begin();
							index != m_oldClients.begin(); ++index) {
		CBaseClientProxy* client = index->first;
		EVENTQUEUE->deleteTimer(index->second);
		EVENTQUEUE->removeHandler(CEvent::kTimer, client);
		EVENTQUEUE->removeHandler(CClientProxy::getDisconnectedEvent(), client);
		delete client;
	}

	// disable and disconnect primary client
	m_primaryClient->disable();
	removeClient(m_primaryClient);
}

bool
CServer::setConfig(const CConfig& config)
{
	// refuse configuration if it doesn't include the primary screen
	if (!config.isScreen(m_primaryClient->getName())) {
		return false;
	}

	// close clients that are connected but being dropped from the
	// configuration.
	closeClients(config);

	// cut over
	m_config = config;
	processOptions();

	// tell primary screen about reconfiguration
	m_primaryClient->reconfigure(getActivePrimarySides());

	// tell all (connected) clients about current options
	for (CClientList::const_iterator index = m_clients.begin();
								index != m_clients.end(); ++index) {
		CBaseClientProxy* client = index->second;
		sendOptions(client);
	}

	return true;
}

void
CServer::adoptClient(CBaseClientProxy* client)
{
	assert(client != NULL);

	// watch for client disconnection
	EVENTQUEUE->adoptHandler(CClientProxy::getDisconnectedEvent(), client,
							new TMethodEventJob<CServer>(this,
								&CServer::handleClientDisconnected, client));

	// name must be in our configuration
	if (!m_config.isScreen(client->getName())) {
		LOG((CLOG_WARN "a client with name \"%s\" is not in the map", client->getName().c_str()));
		closeClient(client, kMsgEUnknown);
		return;
	}

	// add client to client list
	if (!addClient(client)) {
		// can only have one screen with a given name at any given time
		LOG((CLOG_WARN "a client with name \"%s\" is already connected", getName(client).c_str()));
		closeClient(client, kMsgEBusy);
		return;
	}
	LOG((CLOG_NOTE "client \"%s\" has connected", getName(client).c_str()));

	// send configuration options to client
	sendOptions(client);

	// send notification
	CServer::CScreenConnectedInfo* info =
		CServer::CScreenConnectedInfo::alloc(getName(client));
	EVENTQUEUE->addEvent(CEvent(CServer::getConnectedEvent(),
								m_primaryClient->getEventTarget(), info));
}

void
CServer::disconnect()
{
	// close all secondary clients
	if (m_clients.size() > 1 || !m_oldClients.empty()) {
		CConfig emptyConfig;
		closeClients(emptyConfig);
	}
	else {
		EVENTQUEUE->addEvent(CEvent(getDisconnectedEvent(), this));
	}
}

UInt32
CServer::getNumClients() const
{
	return m_clients.size();
}

void
CServer::getClients(std::vector<CStringSYN>& list) const
{
	list.clear();
	for (CClientList::const_iterator index = m_clients.begin();
							index != m_clients.end(); ++index) {
		list.push_back(index->first);
	}
}

CEvent::Type
CServer::getErrorEvent()
{
	return CEvent::registerTypeOnce(s_errorEvent,
							"CServer::error");
}

CEvent::Type
CServer::getConnectedEvent()
{
	return CEvent::registerTypeOnce(s_connectedEvent,
							"CServer::connected");
}

CEvent::Type
CServer::getDisconnectedEvent()
{
	return CEvent::registerTypeOnce(s_disconnectedEvent,
							"CServer::disconnected");
}


CStringSYN
CServer::getName(const CBaseClientProxy* client) const
{
	CStringSYN name = m_config.getCanonicalName(client->getName());
	if (name.empty()) {
		name = client->getName();
	}
	return name;
}

BYTE CServer::getSwitch(CBaseClientProxy* dst) const
{
const CConfig::CScreenOptions* options =
						m_config.getOptions(getName(dst));

for (CConfig::CScreenOptions::const_iterator index = options->begin();
								index != options->end(); ++index) {
		const OptionID id       = index->first;
		const OptionValue value = index->second;
		if (id == kOptionKMSwitch) {
			switch(value){
		case 1:
			return '1';
		case 2:
			return '2';
		case 3:
			return '3';
		case 4:
			return '4';
		default:
			return '0';
			}
		}
	}
return '0';
}

int CServer::sendByte(CBaseClientProxy* dst)
{
	//preprocessor block for user defined serial port path
#if SYSAPI_UNIX
	extern const char* dev;
	dev = m_config.m_serialPort.c_str();
#endif
	assert(uploaderComPort.RS232Open()== CRS232::RS232_OK);
	unsigned char ASCIICode = getSwitch(dst);
	unsigned long byteSize = 1;
	uploaderComPort.RS232Send(&ASCIICode,&byteSize);
	uploaderComPort.RS232Close();
	LOG((CLOG_INFO "sent RS232 command to switch to %d", (int)ASCIICode - 48));
	return 0;
}


UInt32
CServer::getActivePrimarySides() const
{
	UInt32 sides = 0;
	if (hasAnyNeighbor(m_primaryClient, kLeft)) {
		sides |= kLeftMask;
	}
	if (hasAnyNeighbor(m_primaryClient, kRight)) {
		sides |= kRightMask;
	}
	if (hasAnyNeighbor(m_primaryClient, kTop)) {
		sides |= kTopMask;
	}
	if (hasAnyNeighbor(m_primaryClient, kBottom)) {
		sides |= kBottomMask;
	}
	return sides;
}

bool
CServer::isLockedToScreen() const
{
	// locked if primary says we're locked
	if (m_primaryClient->isLockedToScreen()) {
		return true;
	}

	// not locked
	return false;
}

SInt32
CServer::getJumpZoneSize(CBaseClientProxy* client) const
{
	if (client == m_primaryClient) {
		return m_primaryClient->getJumpZoneSize();
	}
	else {
		return 0;
	}
}

void
CServer::switchScreen(CBaseClientProxy* dst,
				SInt32 x, SInt32 y)
{
	// prevent rogue mouse motions interfering with switch
	lockMouseTimer.reset();
	lockMouseTimer.start();
	assert(dst != NULL);
#ifndef NDEBUG
{
	// put definitions back here
	SInt32 dx, dy, dw, dh;
	dst->getShape(dx, dy, dw, dh);
	assert(x >= dx && y >= dy && x < dx + dw && y < dy + dh);
}
#endif
	assert(m_active != NULL);
	
	LOG((CLOG_INFO "switch from \"%s\" to \"%s\" at %d,%d", getName(m_active).c_str(), getName(dst).c_str(), x, y));

	CServer::sendByte(dst);
		
	// stop waiting to switch
	stopSwitch();

	// record new position
	m_x       = x;
	m_y       = y;
	m_xDelta  = 0;
	m_yDelta  = 0;
	m_xDelta2 = 0;
	m_yDelta2 = 0;

	// leave active screen
	if (!m_active->leave()) {
		// cannot leave screen
		LOG((CLOG_WARN "can't leave screen"));
		return;
	}
	// cut over
	m_active = dst;
	// enter new screen
	m_active->enter(x, y);
	m_active->m_hasLeft = false;
}

float
CServer::mapToFraction(CBaseClientProxy* client,
				EDirection dir, SInt32 x, SInt32 y) const
{
	SInt32 sx, sy, sw, sh;
	client->getShape(sx, sy, sw, sh);
	switch (dir) {
	case kLeft:
	case kRight:
		return static_cast<float>(y - sy + 0.5f) / static_cast<float>(sh);

	case kTop:
	case kBottom:
		return static_cast<float>(x - sx + 0.5f) / static_cast<float>(sw);

	case kNoDirection:
		assert(0 && "bad direction");
		break;
	}
	return 0.0f;
}

void
CServer::mapToPixel(CBaseClientProxy* client,
				EDirection dir, float f, SInt32& x, SInt32& y) const
{
	SInt32 sx, sy, sw, sh;
	client->getShape(sx, sy, sw, sh);
	switch (dir) {
	case kLeft:
	case kRight:
		y = static_cast<SInt32>(f * sh) + sy;
		break;

	case kTop:
	case kBottom:
		x = static_cast<SInt32>(f * sw) + sx;
		break;

	case kNoDirection:
		assert(0 && "bad direction");
		break;
	}
}

bool
CServer::hasAnyNeighbor(CBaseClientProxy* client, EDirection dir) const
{
	assert(client != NULL);

	return m_config.hasNeighbor(getName(client), dir);
}

CBaseClientProxy*
CServer::getNeighbor(CBaseClientProxy* src,
				EDirection dir, SInt32& x, SInt32& y) const
{
	// note -- must be locked on entry

	assert(src != NULL);

	// get source screen name
	CStringSYN srcName = getName(src);
	assert(!srcName.empty());
	LOG((CLOG_DEBUG2 "find neighbor on %s of \"%s\"", CConfig::dirName(dir), srcName.c_str()));

	// convert position to fraction
	float t = mapToFraction(src, dir, x, y);

	// search for the closest neighbor that exists in direction dir
	float tTmp;
	for (;;) {
		CStringSYN dstName(m_config.getNeighbor(srcName, dir, t, &tTmp));

		// if nothing in that direction then return NULL. if the
		// destination is the source then we can make no more
		// progress in this direction.  since we haven't found a
		// connected neighbor we return NULL.
		if (dstName.empty()) {
			LOG((CLOG_DEBUG2 "no neighbor on %s of \"%s\"", CConfig::dirName(dir), srcName.c_str()));
			return NULL;
		}

		// look up neighbor cell.  if the screen is connected and
		// ready then we can stop.
		CClientList::const_iterator index = m_clients.find(dstName);
		if (index != m_clients.end()) {
			LOG((CLOG_DEBUG2 "\"%s\" is on %s of \"%s\" at %f", dstName.c_str(), CConfig::dirName(dir), srcName.c_str(), t));
			mapToPixel(index->second, dir, tTmp, x, y);
			return index->second;
		}

		// skip over unconnected screen
		LOG((CLOG_DEBUG2 "ignored \"%s\" on %s of \"%s\"", dstName.c_str(), CConfig::dirName(dir), srcName.c_str()));
		srcName = dstName;

		// use position on skipped screen
		t = tTmp;
	}
}

CBaseClientProxy*
CServer::mapToNeighbor(CBaseClientProxy* src,
				EDirection srcSide, SInt32& x, SInt32& y) const
{
	// note -- must be locked on entry
	assert(src != NULL);

	// get the first neighbor
	CBaseClientProxy* dst = getNeighbor(src, srcSide, x, y);
	if (dst == NULL) {
		return NULL;
	}
	// get the source screen's size
	SInt32 dx, dy, dw, dh;
	CBaseClientProxy* lastGoodScreen = src;
	lastGoodScreen->getShape(dx, dy, dw, dh);

	// find destination screen, adjusting x or y (but not both).  the
	// searches are done in a sort of canonical screen space where
	// the upper-left corner is 0,0 for each screen.  we adjust from
	// actual to canonical position on entry to and from canonical to
	// actual on exit from the search.
	switch (srcSide) {
	case kLeft:
		x -= dx;
		while (dst != NULL) {
			lastGoodScreen = dst;
			lastGoodScreen->getShape(dx, dy, dw, dh);
			x += dw;
			if (x >= 0) {
				break;
			}
			LOG((CLOG_DEBUG2 "skipping over screen %s", getName(dst).c_str()));
			dst = getNeighbor(lastGoodScreen, srcSide, x, y);
		}
		assert(lastGoodScreen != NULL);
		x += dx;
		break;

	case kRight:
		x -= dx;
		while (dst != NULL) {
			x -= dw;
			lastGoodScreen = dst;
			lastGoodScreen->getShape(dx, dy, dw, dh);
			if (x < dw) {
				break;
			}
			LOG((CLOG_DEBUG2 "skipping over screen %s", getName(dst).c_str()));
			dst = getNeighbor(lastGoodScreen, srcSide, x, y);
		}
		assert(lastGoodScreen != NULL);
		x += dx;
		break;

	case kTop:
		y -= dy;
		while (dst != NULL) {
			lastGoodScreen = dst;
			lastGoodScreen->getShape(dx, dy, dw, dh);
			y += dh;
			if (y >= 0) {
				break;
			}
			LOG((CLOG_DEBUG2 "skipping over screen %s", getName(dst).c_str()));
			dst = getNeighbor(lastGoodScreen, srcSide, x, y);
		}
		assert(lastGoodScreen != NULL);
		y += dy;
		break;

	case kBottom:
		y -= dy;
		while (dst != NULL) {
			y -= dh;
			lastGoodScreen = dst;
			lastGoodScreen->getShape(dx, dy, dw, dh);
			if (y < dh) {
				break;
			}
			LOG((CLOG_DEBUG2 "skipping over screen %s", getName(dst).c_str()));
			dst = getNeighbor(lastGoodScreen, srcSide, x, y);
		}
		assert(lastGoodScreen != NULL);
		y += dy;
		break;

	case kNoDirection:
		assert(0 && "bad direction");
		return NULL;
	}

	// save destination screen
	assert(lastGoodScreen != NULL);
	dst = lastGoodScreen;
	// be sure to move in far enough
	// to avoid the jump zone.  if entering a side that doesn't have
	// a neighbor (i.e. an asymmetrical side) then we don't need to
	// move inwards because that side can't provoke a jump.
	avoidJumpZone(dst, srcSide, x, y);
	return dst;
}

void
CServer::avoidJumpZone(CBaseClientProxy* dst,
				EDirection dir, SInt32& x, SInt32& y) const
{
	const CStringSYN dstName(getName(dst));
	SInt32 dx, dy, dw, dh;
	dst->getShape(dx, dy, dw, dh);
	float t = mapToFraction(dst, dir, x, y);
	SInt32 z = getJumpZoneSize(dst);

	// move in far enough to avoid the jump zone.  if entering a side
	// that doesn't have a neighbor (i.e. an asymmetrical side) then we
	// only need to move to the side because that side can't provoke a jump.
	switch (dir) {
	case kLeft:
		if (x > dx + dw - 1 - z) {
			if (!m_config.getNeighbor(dstName, kRight, t, NULL).empty()) {
				x = dx + dw - 1 - z;
			}
			else {
				x = dx + dw - 1;
			}
		}
		break;

	case kRight:
		if (x < dx + z) {
			if (!m_config.getNeighbor(dstName, kLeft, t, NULL).empty()) {
				x = dx + z;
			}
			else {
				x = dx;
			}
		}
		break;

	case kTop:
		if (y > dy + dh - 1 - z) {
			if (!m_config.getNeighbor(dstName, kBottom, t, NULL).empty()) {
			y = dy + dh - 1 - z;
			}
			else {
				y = dy + dh -1;
			}
		}
		break;

	case kBottom:
		if (y < dy + z) {
			if (!m_config.getNeighbor(dstName, kTop, t, NULL).empty()) {
			y = dy + z;
			}
			else {
				y = dy;
			}
		}
		break;

	case kNoDirection:
		assert(0 && "bad direction");
	}
}

bool
CServer::isSwitchOkay(CBaseClientProxy* newScreen,
				EDirection dir, SInt32 x, SInt32 y)
{
	LOG((CLOG_DEBUG1 "try to leave \"%s\" on %s", getName(m_active).c_str(), CConfig::dirName(dir)));

	// is there a neighbor?
	if (newScreen == NULL) {
		// there's no neighbor.  we don't want to switch and we don't
		// want to try to switch later.
		LOG((CLOG_DEBUG1 "no neighbor %s", CConfig::dirName(dir)));
		stopSwitch();
		return false;
	}

	// should we switch or not?
	bool preventSwitch = false;
	bool allowSwitch   = false;

	// note if the switch direction has changed.  save the new
	// direction and screen if so.
	bool isNewDirection  = (dir != m_switchDir);
	if (isNewDirection || m_switchScreen == NULL) {
		m_switchDir    = dir;
		m_switchScreen = newScreen;
	}

	// is this a double tap and do we care?
	if (!allowSwitch && m_switchTwoTapDelay > 0.0) {
		if (isNewDirection ||
			!isSwitchTwoTapStarted() || !shouldSwitchTwoTap()) {
			// tapping a different or new edge or second tap not
			// fast enough.  prepare for second tap.
			preventSwitch = true;
			startSwitchTwoTap(x,y);
		}
		else {
			// got second tap
			allowSwitch = true;
		}
	}

	// if waiting before a switch then prepare to switch later
	if (!allowSwitch && m_switchWaitDelay > 0.0) {
		if (isNewDirection || !isSwitchWaitStarted()) {
			startSwitchWait(x, y);
		}
		preventSwitch = true;
	}

	// are we in a locked corner?  first check if screen has the option set
	// and, if not, check the global options.
	const CConfig::CScreenOptions* options =
						m_config.getOptions(getName(m_active));
	if (options == NULL || options->count(kOptionScreenSwitchCorners) == 0) {
		options = m_config.getOptions("");
	}
	if (options != NULL && options->count(kOptionScreenSwitchCorners) > 0) {
		// get corner mask and size
		CConfig::CScreenOptions::const_iterator i =
			options->find(kOptionScreenSwitchCorners);
		UInt32 corners = static_cast<UInt32>(i->second);
		i = options->find(kOptionScreenSwitchCornerSize);
		SInt32 size = 0;
		if (i != options->end()) {
			size = i->second;
		}

		// see if we're in a locked corner
		if ((getCorner(m_active, x, y, dir, size) & corners) != 0) {
			// yep, no switching
			LOG((CLOG_DEBUG1 "locked in corner"));
			preventSwitch = true;
			stopSwitch();
		}
	}
	// ignore if mouse is locked to screen and don't try to switch later
	if (!preventSwitch && isLockedToScreen()) {
		LOG((CLOG_DEBUG1 "locked to screen"));
		preventSwitch = true;
		stopSwitch();
	}
	return !preventSwitch;
}

void
CServer::noSwitch(SInt32 x, SInt32 y)
{
	armSwitchTwoTap(x, y);
	stopSwitchWait();
}

void
CServer::stopSwitch()
{
	if (m_switchScreen != NULL) {
		m_switchScreen = NULL;
		m_switchDir    = kNoDirection;
		stopSwitchTwoTap();
		stopSwitchWait();
	}
}

void
CServer::startSwitchTwoTap(SInt32 x, SInt32 y)
{
	m_switchTwoTapEngaged = true;
	m_switchTwoTapArmed   = false;
	m_switchTwoTapTimer.reset();
	LOG((CLOG_DEBUG1 "waiting for second tap"));
	armSwitchTwoTap(x, y);
}

void
CServer::armSwitchTwoTap(SInt32 x, SInt32 y)
{
	if (m_switchTwoTapEngaged) {
		if (m_switchTwoTapTimer.getTime() > m_switchTwoTapDelay) {
			// second tap took too long.  disengage.
			LOG((CLOG_DEBUG1 "second tap too late"));
			stopSwitchTwoTap();
		}
		else if (!m_switchTwoTapArmed) {
			// still time for a double tap.  see if we left the tap
			// zone and, if so, arm the two tap.
			SInt32 ax, ay, aw, ah;
			m_active->getShape(ax, ay, aw, ah);
			SInt32 tapZone = m_primaryClient->getJumpZoneSize();
			if (tapZone < m_switchTwoTapZone) {
				tapZone = m_switchTwoTapZone;
			}
			if (x >= ax + tapZone && x < ax + aw - tapZone &&
				y >= ay + tapZone && y < ay + ah - tapZone) {
				// win32 can generate bogus mouse events that appear to
				// move in the opposite direction that the mouse actually
				// moved.  try to ignore that crap here.
				switch (m_switchDir) {
				case kLeft:
					m_switchTwoTapArmed = (m_xDelta > 0 && m_xDelta2 > 0);
					break;

				case kRight:
					m_switchTwoTapArmed = (m_xDelta < 0 && m_xDelta2 < 0);
					break;

				case kTop:
					m_switchTwoTapArmed = (m_yDelta > 0 && m_yDelta2 > 0);
					break;

				case kBottom:
					m_switchTwoTapArmed = (m_yDelta < 0 && m_yDelta2 < 0);
					break;

				default:
					break;
				}
			}
		}
	}
}

void
CServer::stopSwitchTwoTap()
{
	m_switchTwoTapEngaged = false;
	m_switchTwoTapArmed   = false;
}

bool
CServer::isSwitchTwoTapStarted() const
{
	return m_switchTwoTapEngaged;
}

bool
CServer::shouldSwitchTwoTap() const
{
	// this is the second tap if two-tap is armed and this tap
	// came fast enough
	return (m_switchTwoTapArmed &&
			m_switchTwoTapTimer.getTime() <= m_switchTwoTapDelay);
}

SInt8
CServer::isTwoTapActive(void)
{
	if(m_switchTwoTapDelay > 0) {
		return 1;
	}
	else {
		return 0;
	}
}

void
CServer::startSwitchWait(SInt32 x, SInt32 y)
{
	stopSwitchWait();
	m_switchWaitX     = x;
	m_switchWaitY     = y;
	m_switchWaitTimer = EVENTQUEUE->newOneShotTimer(m_switchWaitDelay, this);
	LOG((CLOG_DEBUG1 "waiting to switch"));
}

void
CServer::stopSwitchWait()
{
	if (m_switchWaitTimer != NULL) {
		EVENTQUEUE->deleteTimer(m_switchWaitTimer);
		m_switchWaitTimer = NULL;
	}
}

bool
CServer::isSwitchWaitStarted() const
{
	return (m_switchWaitTimer != NULL);
}

UInt32
CServer::getCorner(CBaseClientProxy* client,
				SInt32 x, SInt32 y, EDirection dir, SInt32 size) const
{
	assert(client != NULL);

	SInt32 ax, ay, aw, ah;
	client->getShape(ax, ay, aw, ah);
	// check direction of switching

	SInt32 xSide, ySide;

	switch (dir) {
		case kLeft:
			xSide = -1;
			ySide = 0;
			break;
		case kRight:
			xSide = 1;
			ySide = 0;
			break;
		case kTop:
			xSide = 0;
			ySide = 1;
			break;
		case kBottom:
			xSide = 0;
			ySide = -1;
			break;
		default:
			xSide = 0;
			ySide = 0;
			return kNoCornerMask;
	}
	
	// if against the left or right then check if y is within size
	if (xSide != 0) {
		if (y < ay + size) {
			return (xSide < 0) ? kTopLeftMask : kTopRightMask;
		}
		else if (y >= ay + ah - size) {
			return (xSide < 0) ? kBottomLeftMask : kBottomRightMask;
		}
	}

	// if against the top or bottom then check if x is within size
	if (ySide != 0) {
		if (x < ax + size) {
			return (ySide < 0) ? kTopLeftMask : kBottomLeftMask;
		}
		else if (x >= ax + aw - size) {
			return (ySide < 0) ? kTopRightMask : kBottomRightMask;
		}
	}

	return kNoCornerMask;
}

void
CServer::sendOptions(CBaseClientProxy* client) const
{
	COptionsList optionsList;

	// look up options for client
	const CConfig::CScreenOptions* options =
						m_config.getOptions(getName(client));
	if (options != NULL) {
		// convert options to a more convenient form for sending
		optionsList.reserve(2 * options->size());
		for (CConfig::CScreenOptions::const_iterator index = options->begin();
									index != options->end(); ++index) {
			optionsList.push_back(index->first);
			optionsList.push_back(static_cast<UInt32>(index->second));
		}
	}

	// look up global options
	options = m_config.getOptions("");
	if (options != NULL) {
		// convert options to a more convenient form for sending
		optionsList.reserve(optionsList.size() + 2 * options->size());
		for (CConfig::CScreenOptions::const_iterator index = options->begin();
									index != options->end(); ++index) {
			optionsList.push_back(index->first);
			optionsList.push_back(static_cast<UInt32>(index->second));
		}
	}

	// send the options
	client->resetOptions();
	client->setOptions(optionsList);
}

void
CServer::processOptions()
{
	const CConfig::CScreenOptions* options = m_config.getOptions("");
	if (options == NULL) {
		return;
	}

	for (CConfig::CScreenOptions::const_iterator index = options->begin();
								index != options->end(); ++index) {
		const OptionID id       = index->first;
		const OptionValue value = index->second;
		if (id == kOptionScreenSwitchDelay) {
			m_switchWaitDelay = 1.0e-3 * static_cast<double>(value);
			if (m_switchWaitDelay < 0.0) {
				m_switchWaitDelay = 0.0;
			}
			stopSwitchWait();
		}
		else if (id == kOptionScreenSwitchTwoTap) {
			m_switchTwoTapDelay = 1.0e-3 * static_cast<double>(value);
			if (m_switchTwoTapDelay < 0.0) {
				m_switchTwoTapDelay = 0.0;
			}
			stopSwitchTwoTap();
		}
	}
}

void
CServer::handleShapeChanged(const CEvent&, void* vclient)
{
	// ignore events from unknown clients
	CBaseClientProxy* client = reinterpret_cast<CBaseClientProxy*>(vclient);
	if (m_clientSet.count(client) == 0) {
		return;
	}

	LOG((CLOG_INFO "screen \"%s\" shape changed", getName(client).c_str()));

	// update jump coordinate
	SInt32 x, y;
	client->getCursorPos(x, y);
	client->setJumpCursorPos(x, y);

	// update the mouse coordinates
	if (client == m_active) {
		m_x = x;
		m_y = y;
	}

	// handle resolution change to primary screen
	if (client == m_primaryClient) {
		if (client == m_active) {
			onMouseMove(m_x, m_y);
		}
	}
}

void
CServer::handleButtonDownEvent(const CEvent& event, void*)
{
	IPlatformScreen::CButtonInfo* info =
		reinterpret_cast<IPlatformScreen::CButtonInfo*>(event.getData());
	onMouseDown(info->m_button);
}

void
CServer::handleButtonUpEvent(const CEvent& event, void*)
{
	IPlatformScreen::CButtonInfo* info =
		reinterpret_cast<IPlatformScreen::CButtonInfo*>(event.getData());
	onMouseUp(info->m_button);
}

void
CServer::handleMotionEvent(const CEvent& event, void*)
{
	IPlatformScreen::CMotionInfo* info =
		reinterpret_cast<IPlatformScreen::CMotionInfo*>(event.getData());
	onMouseMove(info->m_x, info->m_y);
}

void
CServer::handleSwitchWaitTimeout(const CEvent&, void*)
{
	// ignore if mouse is locked to screen
	if (isLockedToScreen()) {
		LOG((CLOG_DEBUG1 "locked to screen"));
		stopSwitch();
		return;
	}

	// switch screen
	switchScreen(m_switchScreen, m_switchWaitX, m_switchWaitY);
}

void
CServer::handleClientDisconnected(const CEvent&, void* vclient)
{
	// client has disconnected.  it might be an old client or an
	// active client.  we don't care so just handle it both ways.
	CBaseClientProxy* client = reinterpret_cast<CBaseClientProxy*>(vclient);
	removeActiveClient(client);
	removeOldClient(client);
	delete client;
}

void
CServer::handleClientCloseTimeout(const CEvent&, void* vclient)
{
	// client took too long to disconnect.  just dump it.
	CBaseClientProxy* client = reinterpret_cast<CBaseClientProxy*>(vclient);
	LOG((CLOG_NOTE "forced disconnection of client \"%s\"", getName(client).c_str()));
	removeOldClient(client);
	delete client;
}

void
CServer::onMouseDown(ButtonID id)
{
	LOG((CLOG_DEBUG1 "onMouseDown id=%d", id));
	assert(m_active != NULL);
}

void
CServer::onMouseUp(ButtonID id)
{
	LOG((CLOG_DEBUG1 "onMouseUp id=%d", id));
	assert(m_active != NULL);
}

bool
CServer::onMouseMove(SInt32 x, SInt32 y)
{
	LOG((CLOG_DEBUG2 "onMouseMove %d,%d", x, y));
	
	if(m_active != m_primaryClient) {

		if(lockMouseTimer.getTime() >= 0.2) {
			lockMouseTimer.stop();
			LOG((CLOG_DEBUG2 "input detected on server"));
			m_active->leave();
			m_primaryClient->enter(x, y);
			m_active = m_primaryClient;
		}
		else {
			LOG((CLOG_DEBUG2 "mouse motion event on server discarded"));
		}
	}

	// save last delta
	m_xDelta2 = m_xDelta;
	m_yDelta2 = m_yDelta;

	// save current delta
	m_xDelta  = x - m_x;
	m_yDelta  = y - m_y;

	// save position
	m_x       = x;
	m_y       = y;

	SInt32 ax, ay, aw, ah;
	m_primaryClient->getShape(ax, ay, aw, ah);
	SInt32 zoneSize = getJumpZoneSize(m_primaryClient);

	// see if we should change screens and clamp to sides
	EDirection dir;
	if (x < ax + zoneSize) {
		x = ax;
		dir = kLeft;
	}
	else if (x >= ax + aw - zoneSize) {
		x = ax + aw - 1;
		dir = kRight;
	}
	else if (y < ay + zoneSize) {
		y = ay;
		dir = kTop;
	}
	else if (y >= ay + ah - zoneSize) {
		y = ay + ah - 1;
		dir = kBottom;
	}
	else {
		// still on local screen
		noSwitch(x, y);
		return false;
	}
	LOG((CLOG_DEBUG1 "reached jumpzone on server- looking for neighbours"));
	
	// get jump destination
	CBaseClientProxy* newScreen = mapToNeighbor(m_active, dir, x, y);
	if (newScreen != NULL) {
	}
	// should we switch or not?
	if (isSwitchOkay(newScreen, dir, x, y)) {
		// switch screen
		switchScreen(newScreen, x, y);
		return true;
	}
	else {
		return false;
	}
}

bool
CServer::addClient(CBaseClientProxy* client)
{
	CStringSYN name = getName(client);
	if (m_clients.count(name) != 0) {
		return false;
	}

	// add event handlers
	EVENTQUEUE->adoptHandler(IScreen::getShapeChangedEvent(),
							client->getEventTarget(),
							new TMethodEventJob<CServer>(this,
								&CServer::handleShapeChanged, client));
	
	// add to list
	m_clientSet.insert(client);
	m_clients.insert(std::make_pair(name, client));
	// initialize client data
	SInt32 x, y;
	client->getCursorPos(x, y);
	client->setJumpCursorPos(x, y);
	// tell primary client about the active sides
	m_primaryClient->reconfigure(getActivePrimarySides());
	return true;
}

bool
CServer::removeClient(CBaseClientProxy* client)
{
	// return false if not in list
	CClientSet::iterator i = m_clientSet.find(client);
	if (i == m_clientSet.end()) {
		return false;
	}

	// remove event handlers
	EVENTQUEUE->removeHandler(IScreen::getShapeChangedEvent(),
							client->getEventTarget());
	
	// remove from list
	m_clients.erase(getName(client));
	m_clientSet.erase(i);

	return true;
}

void
CServer::closeClient(CBaseClientProxy* client, const char* msg)
{
	assert(client != m_primaryClient);
	assert(msg != NULL);

	// send message to client.  this message should cause the client
	// to disconnect.  we add this client to the closed client list
	// and install a timer to remove the client if it doesn't respond
	// quickly enough.  we also remove the client from the active
	// client list since we're not going to listen to it anymore.
	// note that this method also works on clients that are not in
	// the m_clients list.  adoptClient() may call us with such a
	// client.
	LOG((CLOG_NOTE "disconnecting client \"%s\"", getName(client).c_str()));

	// send message
	// FIXME -- avoid type cast (kinda hard, though)
	((CClientProxy*)client)->close(msg);

	// install timer.  wait timeout seconds for client to close.
	double timeout = 5.0;
	CEventQueueTimer* timer = EVENTQUEUE->newOneShotTimer(timeout, NULL);
	EVENTQUEUE->adoptHandler(CEvent::kTimer, timer,
							new TMethodEventJob<CServer>(this,
								&CServer::handleClientCloseTimeout, client));

	// move client to closing list
	removeClient(client);
	m_oldClients.insert(std::make_pair(client, timer));

	// if this client is the active screen then we have to
	// jump off of it
	forceLeaveClient(client);
}

void
CServer::closeClients(const CConfig& config)
{
	// collect the clients that are connected but are being dropped
	// from the configuration (or who's canonical name is changing).
	typedef std::set<CBaseClientProxy*> CRemovedClients;
	CRemovedClients removed;
	for (CClientList::iterator index = m_clients.begin();
								index != m_clients.end(); ++index) {
		if (!config.isCanonicalName(index->first)) {
			removed.insert(index->second);
		}
	}

	// don't close the primary client
	removed.erase(m_primaryClient);

	// now close them.  we collect the list then close in two steps
	// because closeClient() modifies the collection we iterate over.
	for (CRemovedClients::iterator index = removed.begin();
								index != removed.end(); ++index) {
		closeClient(*index, kMsgCClose);
	}
}

void
CServer::removeActiveClient(CBaseClientProxy* client)
{
	if (removeClient(client)) {
		forceLeaveClient(client);
		EVENTQUEUE->removeHandler(CClientProxy::getDisconnectedEvent(), client);
		if (m_clients.size() == 1 && m_oldClients.empty()) {
			EVENTQUEUE->addEvent(CEvent(getDisconnectedEvent(), this));
		}
	}
}

void
CServer::removeOldClient(CBaseClientProxy* client)
{
	COldClients::iterator i = m_oldClients.find(client);
	if (i != m_oldClients.end()) {
		EVENTQUEUE->removeHandler(CClientProxy::getDisconnectedEvent(), client);
		EVENTQUEUE->removeHandler(CEvent::kTimer, i->second);
		EVENTQUEUE->deleteTimer(i->second);
		m_oldClients.erase(i);
		if (m_clients.size() == 1 && m_oldClients.empty()) {
			EVENTQUEUE->addEvent(CEvent(getDisconnectedEvent(), this));
		}
	}
}

void
CServer::forceLeaveClient(CBaseClientProxy* client)
{
	CBaseClientProxy* active = m_active;
	if (active == client) {
		// record new position (center of primary screen)
		m_primaryClient->getCursorCenter(m_x, m_y);

		// stop waiting to switch to this client
		if (active == m_switchScreen) {
			stopSwitch();
		}

		// don't notify active screen since it has probably already
		// disconnected.
		LOG((CLOG_INFO "jump from \"%s\" to \"%s\" at %d,%d", getName(active).c_str(), getName(m_primaryClient).c_str(), m_x, m_y));

		// cut over
		m_active = m_primaryClient;
		m_primaryClient->enter(m_x, m_y);
	}

	// tell primary client about the active sides
	m_primaryClient->reconfigure(getActivePrimarySides());
}


//
// CServer::CScreenConnectedInfo
//

CServer::CScreenConnectedInfo*
CServer::CScreenConnectedInfo::alloc(const CStringSYN& screen)
{
	CScreenConnectedInfo* info =
		(CScreenConnectedInfo*)malloc(sizeof(CScreenConnectedInfo) +
								screen.size());
	strcpy(info->m_screen, screen.c_str());
	return info;
}
